@@ -8,37 +8,44 @@ from future_builtins import filter, map, zip
 import os
 import itertools
 import contextlib
-import abc, collections
+import abc
+import collections
 import lucene
 from java.io import File, StringReader
 from java.lang import Float
 from java.util import Arrays, HashMap, HashSet
-from org.apache.lucene import analysis, codecs, document, index, queries, queryparser, search, store, util
+from org.apache.lucene import analysis, document, index, queries, queryparser, search, store, util
 from org.apache.pylucene.analysis import PythonAnalyzer, PythonTokenFilter
 from org.apache.pylucene.queryparser.classic import PythonQueryParser
-from .queries import Query, BooleanFilter, TermsFilter, SortField, Highlighter, FastVectorHighlighter, SpellChecker, SpellParser
+from .queries import suppress, Query, BooleanFilter, TermsFilter, SortField, Highlighter, FastVectorHighlighter, SpellChecker, SpellParser
 from .documents import Field, Document, Hits, GroupingSearch
 from .spatial import DistanceComparator
 
+
 class Atomic(object):
     "Abstract base class to distinguish singleton values from other iterables."
     __metaclass__ = abc.ABCMeta
+
     @classmethod
     def __subclasshook__(cls, other):
         return not issubclass(other, collections.Iterable) or NotImplemented
+
 for cls in (basestring, analysis.TokenStream, lucene.JArray_byte):
     Atomic.register(cls)
 
+
 class closing(set):
     "Manage lifespan of registered objects, similar to contextlib.closing."
     def __del__(self):
         for obj in self:
             obj.close()
+
     def analyzer(self, analyzer, version=None):
         if analyzer is None:
             analyzer = analysis.standard.StandardAnalyzer(version or util.Version.values()[-1])
             self.add(analyzer)
         return analyzer
+
     def directory(self, directory):
         if directory is None:
             directory = store.RAMDirectory()
@@ -47,16 +54,25 @@ class closing(set):
             directory = store.FSDirectory.open(File(directory))
             self.add(directory)
         return directory
+
     def reader(self, reader):
-        reader = self.directory(reader)
         if isinstance(reader, index.IndexReader):
             reader.incRef()
         elif isinstance(reader, index.IndexWriter):
             reader = index.IndexReader.open(reader, True)
+        elif isinstance(reader, Atomic):
+            reader = index.IndexReader.open(self.directory(reader))
         else:
-            reader = index.IndexReader.open(reader)
+            reader = index.MultiReader(list(map(self.reader, reader)))
         return reader
 
+    @classmethod
+    @contextlib.contextmanager
+    def store(cls, directory):
+        self = cls()
+        yield self.directory(directory)
+
+
 def copy(commit, dest):
     """Copy the index commit to the destination directory.
     Optimized to use hard links if the destination is a file system path.
@@ -75,20 +91,24 @@ def copy(commit, dest):
                 if not os.path.samefile(*paths):
                     raise
 
+
 class TokenStream(analysis.TokenStream):
     "TokenStream mixin with support for iteration and attributes cached as properties."
     def __iter__(self):
         self.reset()
         return self
+
     def next(self):
         if self.incrementToken():
             return self
         raise StopIteration
+
     def __getattr__(self, name):
         cls = getattr(analysis.tokenattributes, name + 'Attribute').class_
         attr = self.getAttribute(cls) if self.hasAttribute(cls) else self.addAttribute(cls)
         setattr(self, name, attr)
         return attr
+
     @property
     def offset(self):
         "Start and stop character offset."
@@ -96,6 +116,7 @@ class TokenStream(analysis.TokenStream):
     @offset.setter
     def offset(self, item):
         self.Offset.setOffset(*item)
+
     @property
     def payload(self):
         "Payload bytes."
@@ -104,6 +125,7 @@ class TokenStream(analysis.TokenStream):
     @payload.setter
     def payload(self, data):
         self.Payload.payload = util.BytesRef(data)
+
     @property
     def positionIncrement(self):
         "Position relative to the previous token."
@@ -111,6 +133,7 @@ class TokenStream(analysis.TokenStream):
     @positionIncrement.setter
     def positionIncrement(self, index):
         self.PositionIncrement.positionIncrement = index
+
     @property
     def term(self):
         "Term text."
@@ -119,6 +142,7 @@ class TokenStream(analysis.TokenStream):
     def term(self, text):
         self.CharTerm.setEmpty()
         self.CharTerm.append(text)
+
     @property
     def type(self):
         "Lexical type."
@@ -127,6 +151,7 @@ class TokenStream(analysis.TokenStream):
     def type(self, text):
         self.Type.setType(text)
 
+
 class TokenFilter(PythonTokenFilter, TokenStream):
     """Create an iterable lucene TokenFilter from a TokenStream.
     Subclass and override :meth:`incrementToken` or :meth:`setattrs`.
@@ -134,14 +159,17 @@ class TokenFilter(PythonTokenFilter, TokenStream):
     def __init__(self, input):
         PythonTokenFilter.__init__(self, input)
         self.input = input
+
     def incrementToken(self):
         "Advance to next token and return whether the stream is not empty."
         result = self.input.incrementToken()
         self.setattrs()
         return result
+
     def setattrs(self):
         "Customize current token."
 
+
 class Analyzer(PythonAnalyzer):
     """Return a lucene Analyzer which chains together a tokenizer and filters.
     
@@ -151,16 +179,20 @@ class Analyzer(PythonAnalyzer):
     def __init__(self, tokenizer, *filters):
         PythonAnalyzer.__init__(self)
         self.tokenizer, self.filters = tokenizer, filters
+
     def components(self, field, reader):
         source = tokens = self.tokenizer.tokenStream(field, reader) if isinstance(self.tokenizer, analysis.Analyzer) else self.tokenizer(reader)
         for filter in self.filters:
             tokens = filter(tokens)
         return source, tokens
+
     def createComponents(self, field, reader):
         return analysis.Analyzer.TokenStreamComponents(*self.components(field, reader))
+
     def tokens(self, text, field=None):
         "Return lucene TokenStream from text."
         return self.components(field, StringReader(text))[1]
+
     def parse(self, query, field='', op='', version='', parser=None, **attrs):
         """Return parsed lucene Query.
         
@@ -193,6 +225,7 @@ class Analyzer(PythonAnalyzer):
             if isinstance(parser, PythonQueryParser):
                 parser.finalize()
 
+
 class IndexReader(object):
     """Delegated lucene IndexReader, with a mapping interface of ids to document objects.
     
@@ -200,39 +233,50 @@ class IndexReader(object):
     """
     def __init__(self, reader):
         self.indexReader = reader
+
     def __getattr__(self, name):
         if name == 'indexReader':
             raise AttributeError(name)
         return getattr(index.DirectoryReader.cast_(self.indexReader), name)
+
     def __len__(self):
         return self.numDocs()
+
     def __contains__(self, id):
         bits = index.MultiFields.getLiveDocs(self.indexReader)
         return (0 <= id < self.maxDoc()) and (not bits or bits.get(id))
+
     def __iter__(self):
         ids = xrange(self.maxDoc())
         bits = index.MultiFields.getLiveDocs(self.indexReader)
         return filter(bits.get, ids) if bits else iter(ids)
+
     @property
     def directory(self):
         "reader's lucene Directory"
         return self.__getattr__('directory')()
+
     @property
     def path(self):
         "FSDirectory path"
         return store.FSDirectory.cast_(self.directory).directory.path
+
     @property
     def timestamp(self):
         "timestamp of reader's last commit"
-        return os.path.getmtime(os.path.join(self.path, self.indexCommit.segmentsFileName))
+        directory = store.FSDirectory.cast_(self.directory).directory
+        return File(directory, self.indexCommit.segmentsFileName).lastModified() * 0.001
+
     @property
     def readers(self):
         "segment readers"
         return (index.SegmentReader.cast_(context.reader()) for context in self.leaves())
+
     @property
     def segments(self):
         "segment filenames with document counts"
         return dict((reader.segmentName, reader.numDocs()) for reader in self.readers)
+
     def copy(self, dest, query=None, exclude=None, merge=0):
         """Copy the index to the destination directory.
         Optimized to use hard links if the destination is a file system path.
@@ -253,6 +297,7 @@ class IndexReader(object):
             if merge:
                 writer.forceMerge(merge)
             return len(writer)
+
     def names(self, **attrs):
         """Return field names, given option description.
         
@@ -260,6 +305,7 @@ class IndexReader(object):
         """
         fieldinfos = index.MultiFields.getMergedFieldInfos(self.indexReader).iterator()
         return [fieldinfo.name for fieldinfo in fieldinfos if all(getattr(fieldinfo, name) == attrs[name] for name in attrs)]
+
     def terms(self, name, value='', stop=None, counts=False, distance=0):
         """Generate a slice of term values, optionally with frequency counts.
         Supports a range of terms, wildcard terms, or fuzzy terms.
@@ -281,6 +327,7 @@ class IndexReader(object):
             for bytesref in termenum:
                 text = bytesref.utf8ToString()
                 yield (text, termenum.docFreq()) if counts else text
+
     def numbers(self, name, step=0, type=int, counts=False):
         """Generate decoded numeric term values, optionally with frequency counts.
         
@@ -296,12 +343,14 @@ class IndexReader(object):
         for bytesref in termenum:
             value = convert(decode(bytesref))
             yield (value, termenum.docFreq()) if counts else value
+
     def docs(self, name, value, counts=False):
         "Generate doc ids which contain given term, optionally with frequency counts."
         docsenum = index.MultiFields.getTermDocsEnum(self.indexReader, index.MultiFields.getLiveDocs(self.indexReader), name, util.BytesRef(value))
         if docsenum:
             for doc in iter(docsenum.nextDoc, index.DocsEnum.NO_MORE_DOCS):
                 yield (doc, docsenum.freq()) if counts else doc
+
     def positions(self, name, value, payloads=False):
         "Generate doc ids and positions which contain given term, optionally only with payloads."
         docsenum = index.MultiFields.getTermPositionsEnum(self.indexReader, index.MultiFields.getLiveDocs(self.indexReader), name, util.BytesRef(value))
@@ -312,6 +361,7 @@ class IndexReader(object):
                     yield doc, [(position, docsenum.payload.utf8ToString()) for position in positions if docsenum.payload]
                 else:
                     yield doc, list(positions)
+
     def spans(self, query, positions=False, payloads=False):
         """Generate docs with occurrence counts for a span query.
         
@@ -325,23 +375,26 @@ class IndexReader(object):
             for doc, spans in itertools.groupby(itertools.takewhile(search.spans.Spans.next, spans), key=search.spans.Spans.doc):
                 doc += offset
                 if payloads:
-                    yield doc, [(span.start(), span.end(), [lucene.JArray_byte.cast_(data).string_ for data in span.payload]) \
-                        for span in spans if span.payloadAvailable]
+                    yield doc, [(span.start(), span.end(), [lucene.JArray_byte.cast_(data).string_ for data in span.payload])
+                                for span in spans if span.payloadAvailable]
                 elif positions:
                     yield doc, [(span.start(), span.end()) for span in spans]
                 else:
                     yield doc, sum(1 for span in spans)
             offset += reader.maxDoc()
+
     def vector(self, id, field):
         terms = self.getTermVector(id, field)
         if terms:
             termenum = terms.iterator(None)
             for bytesref in util.BytesRefIterator.cast_(termenum):
                 yield bytesref.utf8ToString(), termenum
+
     def termvector(self, id, field, counts=False):
         "Generate terms for given doc id and field, optionally with frequency counts."
         for term, termenum in self.vector(id, field):
             yield (term, int(termenum.totalTermFreq())) if counts else term
+
     def positionvector(self, id, field, offsets=False):
         "Generate terms and positions for given doc id and field, optionally with character offsets."
         for term, termenum in self.vector(id, field):
@@ -352,6 +405,7 @@ class IndexReader(object):
                 yield term, [(docsenum.startOffset(), docsenum.endOffset()) for position in positions]
             else:
                 yield term, list(positions)
+
     def morelikethis(self, doc, *fields, **attrs):
         """Return MoreLikeThis query for document.
         
@@ -365,6 +419,7 @@ class IndexReader(object):
             setattr(mlt, name, value)
         return mlt.like(StringReader(doc), '') if isinstance(doc, basestring) else mlt.like(doc)
 
+
 class IndexSearcher(search.IndexSearcher, IndexReader):
     """Inherited lucene IndexSearcher, with a mixed-in IndexReader.
     
@@ -377,18 +432,23 @@ class IndexSearcher(search.IndexSearcher, IndexReader):
         self.analyzer = self.shared.analyzer(analyzer)
         self.filters, self.sorters, self.spellcheckers = {}, {}, {}
         self.termsfilters = set()
+
     @classmethod
     def load(cls, directory, analyzer=None):
         "Open `IndexSearcher`_ with a lucene RAMDirectory, loading index into memory."
-        ref = closing()
-        directory = ref.directory(directory)
-        directory = store.RAMDirectory(directory, store.IOContext.DEFAULT)
+        with closing.store(directory) as directory:
+            directory = store.RAMDirectory(directory, store.IOContext.DEFAULT)
         self = cls(directory, analyzer)
         self.shared.add(self.directory)
         return self
+
     def __del__(self):
         if hash(self):
             self.decRef()
+
+    def openIfChanged(self):
+        return index.DirectoryReader.openIfChanged(index.DirectoryReader.cast_(self.indexReader))
+
     def reopen(self, filters=False, sorters=False, spellcheckers=False):
         """Return current `IndexSearcher`_, only creating a new one if necessary.
         Any registered :attr:`termsfilters` are also refreshed.
@@ -397,17 +457,14 @@ class IndexSearcher(search.IndexSearcher, IndexReader):
         :param sorters: refresh cached :attr:`sorters` with associated parsers
         :param spellcheckers: refresh cached :attr:`spellcheckers`
         """
-        try:
-            reader = index.DirectoryReader.openIfChanged(index.DirectoryReader.cast_(self.indexReader))
-        except TypeError:
-            readers = list(map(index.DirectoryReader.openIfChanged, self.indexReaders))
-            reader = index.MultiReader([new or old for new, old in zip(readers, self.indexReaders)]) if any(readers) else None
+        reader = self.openIfChanged()
         if reader is None:
             return self
         other = type(self)(reader, self.analyzer)
         other.decRef()
         other.shared = self.shared
         other.filters.update((key, value if isinstance(value, search.Filter) else dict(value)) for key, value in self.filters.items())
+        other.termsfilters.update(self.termsfilters)
         for termsfilter in self.termsfilters:
             termsfilter.refresh(other)
         if filters:
@@ -422,23 +479,28 @@ class IndexSearcher(search.IndexSearcher, IndexReader):
         else:
             other.spellcheckers = dict(self.spellcheckers)
         return other
+
     def __getitem__(self, id):
         return Document(self.doc(id))
+
     def get(self, id, *fields):
         "Return `Document`_ with only selected fields loaded."
         return Document(self.document(id, HashSet(Arrays.asList(fields))))
+
     def parse(self, query, spellcheck=False, **kwargs):
         if isinstance(query, search.Query):
             return query
         if spellcheck:
             kwargs['parser'], kwargs['searcher'] = SpellParser, self
         return Analyzer.__dict__['parse'](self.analyzer, query, **kwargs)
+
     def highlighter(self, query, field, **kwargs):
         "Return `Highlighter`_ or if applicable `FastVectorHighlighter`_ specific to searcher and query."
         query = self.parse(query, field=field)
         fieldinfo = index.MultiFields.getMergedFieldInfos(self.indexReader).fieldInfo(field)
         vector = fieldinfo and fieldinfo.hasVectors()
         return (FastVectorHighlighter if vector else Highlighter)(self, query, field, **kwargs)
+
     def count(self, *query, **options):
         """Return number of hits for given query or term.
         
@@ -451,6 +513,7 @@ class IndexSearcher(search.IndexSearcher, IndexReader):
         collector = search.TotalHitCountCollector()
         search.IndexSearcher.search(self, query, options.get('filter'), collector)
         return collector.totalHits
+
     def collector(self, query, count=None, sort=None, reverse=False, scores=False, maxscore=False):
         inorder = not self.createNormalizedWeight(query).scoresDocsOutOfOrder()
         if count is None:
@@ -463,6 +526,7 @@ class IndexSearcher(search.IndexSearcher, IndexReader):
         if not isinstance(sort, search.Sort):
             sort = search.Sort(sort)
         return search.TopFieldCollector.create(sort, count, True, scores, maxscore, inorder)
+
     def search(self, query=None, filter=None, count=None, sort=None, reverse=False, scores=False, maxscore=False, timeout=None, **parser):
         """Run query and return `Hits`_.
         
@@ -482,19 +546,18 @@ class IndexSearcher(search.IndexSearcher, IndexReader):
         cache = collector = self.collector(query, count, sort, reverse, scores, maxscore)
         counter = search.TimeLimitingCollector.getGlobalCounter()
         results = collector if timeout is None else search.TimeLimitingCollector(collector, counter, long(timeout * 1000))
-        try:
+        with suppress(search.TimeLimitingCollector.TimeExceededException):
             search.IndexSearcher.search(self, query, filter, results)
-        except lucene.JavaError as timeout:
-            if not search.TimeLimitingCollector.TimeExceededException.instance_(timeout.getJavaException()):
-                raise
+            timeout = None
         if isinstance(cache, search.CachingCollector):
             collector = search.TotalHitCountCollector()
             cache.replay(collector)
             collector = self.collector(query, collector.totalHits or 1, sort, reverse, scores, maxscore)
             cache.replay(collector)
         topdocs = collector.topDocs()
-        stats = (topdocs.totalHits, topdocs.maxScore) * (not isinstance(timeout, lucene.JavaError))
+        stats = (topdocs.totalHits, topdocs.maxScore) * (timeout is None)
         return Hits(self, topdocs.scoreDocs, *stats)
+
     def facets(self, query, *keys):
         """Return mapping of document counts for the intersection with each facet.
         
@@ -523,13 +586,16 @@ class IndexSearcher(search.IndexSearcher, IndexReader):
                 name, value = key
                 counts[name][value] = self.count(filter=BooleanFilter.all(query, self.filters[name][value]))
         return dict(counts)
+
     def groupby(self, field, query, filter=None, count=None, start=0, **attrs):
         "Return `Hits`_ grouped by field using a `GroupingSearch`_."
         return GroupingSearch(field, **attrs).search(self, self.parse(query), filter, count, start)
+
     def sorter(self, field, type='string', parser=None, reverse=False):
         "Return `SortField`_ with cached attributes if available."
         sorter = self.sorters.get(field, SortField(field, type, parser, reverse))
         return sorter if sorter.reverse == reverse else SortField(sorter.field, sorter.typename, sorter.parser, reverse)
+
     def comparator(self, field, type='string', parser=None, multi=False):
         """Return cache of field values suitable for sorting, using a cached `SortField`_ if available.
         Parsing values into an array is memory optimized.
@@ -542,19 +608,23 @@ class IndexSearcher(search.IndexSearcher, IndexReader):
         :param multi: retrieve multi-valued string terms as a tuple
         """
         return self.sorter(field, type, parser).comparator(self, multi)
+
     def distances(self, lng, lat, lngfield, latfield):
         "Return distance comparator computed from cached lat/lng fields."
         arrays = (self.comparator(field, 'double') for field in (lngfield, latfield))
         return DistanceComparator(lng, lat, *arrays)
+
     def spellchecker(self, field):
         "Return and cache spellchecker for given field."
         try:
             return self.spellcheckers[field]
         except KeyError:
             return self.spellcheckers.setdefault(field, SpellChecker(self.terms(field, counts=True)))
+
     def suggest(self, field, prefix, count=None):
         "Return ordered suggested words for prefix."
         return self.spellchecker(field).suggest(prefix, count)
+
     def correct(self, field, text, distance=2):
         """Generate potential words ordered by increasing edit distance and decreasing frequency.
         For optimal performance only iterate the required slice size of corrections.
@@ -562,6 +632,7 @@ class IndexSearcher(search.IndexSearcher, IndexReader):
         :param distance: the maximum edit distance to consider for enumeration
         """
         return itertools.chain.from_iterable(itertools.islice(self.spellchecker(field).correct(text), distance + 1))
+
     def match(self, document, *queries):
         "Generate scores for all queries against a given document mapping."
         searcher = index.memory.MemoryIndex()
@@ -574,6 +645,18 @@ class IndexSearcher(search.IndexSearcher, IndexReader):
         for query in queries:
             yield searcher.search(self.parse(query))
 
+    def termsfilter(self, field, values=()):
+        """Return registered `TermsFilter`_, which will be refreshed whenever the searcher is reopened.
+        
+        .. versionadded:: 1.7
+        .. note:: This interface is experimental and might change in incompatible ways in the next release.
+        """
+        termsfilter = TermsFilter(field, values)
+        termsfilter.refresh(self)
+        self.termsfilters.add(termsfilter)
+        return termsfilter
+
+
 class MultiSearcher(IndexSearcher):
     """IndexSearcher with underlying lucene MultiReader.
     
@@ -581,21 +664,23 @@ class MultiSearcher(IndexSearcher):
     :param analyzer: lucene Analyzer, default StandardAnalyzer
     """
     def __init__(self, reader, analyzer=None):
-        shared = closing()
-        if not index.MultiReader.instance_(reader):
-            reader = index.MultiReader(list(map(shared.reader, reader)))
-            ref = closing([reader])
         IndexSearcher.__init__(self, reader, analyzer)
-        self.shared.update(shared)
-        shared.clear()
         self.indexReaders = [index.DirectoryReader.cast_(context.reader()) for context in self.context.children()]
         self.version = sum(reader.version for reader in self.indexReaders)
+
     def __getattr__(self, name):
         return getattr(index.MultiReader.cast_(self.indexReader), name)
+
+    def openIfChanged(self):
+        readers = list(map(index.DirectoryReader.openIfChanged, self.indexReaders))
+        if any(readers):
+            return index.MultiReader([new or old.incRef() or old for new, old in zip(readers, self.indexReaders)])
+
     @property
     def timestamp(self):
         return max(IndexReader(reader).timestamp for reader in self.indexReaders)
 
+
 class IndexWriter(index.IndexWriter):
     """Inherited lucene IndexWriter.
     Supports setting fields parameters explicitly, so documents can be represented as dictionaries.
@@ -608,6 +693,7 @@ class IndexWriter(index.IndexWriter):
     """
     __len__ = index.IndexWriter.numDocs
     parse = IndexSearcher.__dict__['parse']
+
     def __init__(self, directory=None, mode='a', analyzer=None, version=None, **attrs):
         self.shared = closing()
         if version is None:
@@ -616,29 +702,29 @@ class IndexWriter(index.IndexWriter):
         config.openMode = index.IndexWriterConfig.OpenMode.values()['wra'.index(mode)]
         for name, value in attrs.items():
             setattr(config, name, value)
-        config.indexDeletionPolicy = index.SnapshotDeletionPolicy(config.indexDeletionPolicy)
+        self.policy = config.indexDeletionPolicy = index.SnapshotDeletionPolicy(config.indexDeletionPolicy)
         index.IndexWriter.__init__(self, self.shared.directory(directory), config)
-        self.policy = index.SnapshotDeletionPolicy.cast_(self.config.indexDeletionPolicy)
         self.fields = {}
+
     def __del__(self):
         if hash(self):
             self.close()
+
     @classmethod
     def check(cls, directory, fix=False):
         "Check and optionally fix unlocked index, returning lucene CheckIndex.Status."
-        ref = closing()
-        directory = ref.directory(directory)
-        checkindex = index.CheckIndex(directory)
-        lock = directory.makeLock(cls.WRITE_LOCK_NAME)
-        assert lock.obtain(), "index must not be opened by any writer"
-        try:
-            status = checkindex.checkIndex()
-            if fix:
-                args = [codecs.Codec.getDefault()] if lucene.VERSION < '4.5' else []
-                checkindex.fixIndex(status, *args)
-        finally:
-            lock.close() if hasattr(lock, 'close') else lock.release()
+        with closing.store(directory) as directory:
+            checkindex = index.CheckIndex(directory)
+            lock = directory.makeLock(cls.WRITE_LOCK_NAME)
+            assert lock.obtain(), "index must not be opened by any writer"
+            try:
+                status = checkindex.checkIndex()
+                if fix:
+                    checkindex.fixIndex(status)
+            finally:
+                lock.close() if hasattr(lock, 'close') else lock.release()
         return status
+
     def set(self, name, cls=Field, **settings):
         """Assign settings to field name.
         
@@ -647,6 +733,7 @@ class IndexWriter(index.IndexWriter):
         :param settings: stored, indexed, etc. options compatible with `Field`_
         """
         self.fields[name] = cls(name, **settings)
+
     def document(self, items=(), **terms):
         "Return lucene Document from mapping of field names to one or multiple values."
         doc = document.Document()
@@ -656,13 +743,27 @@ class IndexWriter(index.IndexWriter):
             for field in self.fields[name].items(*values):
                 doc.add(field)
         return doc
+
     def add(self, document=(), **terms):
         "Add :meth:`document` to index with optional boost."
         self.addDocument(self.document(document, **terms))
+
     def update(self, name, value='', document=(), **terms):
-        "Atomically delete documents which match given term and add the new :meth:`document` with optional boost."
-        doc = self.document(document, **terms)
-        self.updateDocument(index.Term(name, *[value] if value else doc.getValues(name)), doc)
+        """Atomically delete documents which match given term and add the new :meth:`document`.
+        
+        .. versionchanged:: 1.7 update in-place if only DocValues are given; lucene >= 4.8 required for binary values
+        """
+        document = dict(document, **terms)
+        term = index.Term(name, value or document[name])
+        if all(self.fields[name].docValueType() for name in document):
+            for name, value in document.items():
+                if isinstance(value, int):
+                    self.updateNumericDocValue(term, name, long(value))
+                else:
+                    self.updateBinaryDocValue(term, name, util.BytesRef(value))
+        else:
+            self.updateDocument(term, self.document(document))
+
     def delete(self, *query, **options):
         """Remove documents which match given query or term.
         
@@ -671,17 +772,18 @@ class IndexWriter(index.IndexWriter):
         """
         parse = self.parse if len(query) == 1 else index.Term
         self.deleteDocuments(parse(*query, **options))
+
     def __iadd__(self, directory):
         "Add directory (or reader, searcher, writer) to index."
-        ref = closing()
-        directory = ref.directory(directory)
-        self.addIndexes([directory if isinstance(directory, store.Directory) else directory.directory])
+        with closing.store(getattr(directory, 'directory', directory)) as directory:
+            self.addIndexes([directory])
         return self
+
     @contextlib.contextmanager
     def snapshot(self):
         """Return context manager of an index commit snapshot.
         
-        .. versionchanged:: 1.4 lucene 4.4 identifies snapshots by commit generation
+        .. versionchanged:: 1.4 lucene identifies snapshots by commit generation
         """
         commit = self.policy.snapshot()
         try:
@@ -689,6 +791,7 @@ class IndexWriter(index.IndexWriter):
         finally:
             self.policy.release(commit)
 
+
 class Indexer(IndexWriter):
     """An all-purpose interface to an index.
     Creates an `IndexWriter`_ with a delegated `IndexSearcher`_.
@@ -700,19 +803,25 @@ class Indexer(IndexWriter):
         IndexWriter.commit(self)
         self.nrt = nrt
         self.indexSearcher = IndexSearcher(self if nrt else self.directory, self.analyzer)
+
     def __getattr__(self, name):
         if name == 'indexSearcher':
             raise AttributeError(name)
         return getattr(self.indexSearcher, name)
+
     def __contains__(self, id):
         return id in self.indexSearcher
+
     def __iter__(self):
         return iter(self.indexSearcher)
+
     def __getitem__(self, id):
         return self.indexSearcher[id]
+
     def refresh(self, **caches):
         "Store refreshed searcher with :meth:`IndexSearcher.reopen` caches."
         self.indexSearcher = self.indexSearcher.reopen(**caches)
+
     def commit(self, merge=False, **caches):
         """Commit writes and :meth:`refresh` searcher.
         
@@ -727,6 +836,7 @@ class Indexer(IndexWriter):
             IndexWriter.commit(self)
         self.refresh(**caches)
 
+
 class ParallelIndexer(Indexer):
     """Indexer which tracks a unique identifying field.
     Handles atomic updates of rapidly changing fields, managing :attr:`termsfilters`.
@@ -737,6 +847,7 @@ class ParallelIndexer(Indexer):
         self.field = field
         self.set(field, tokenized=False, omitNorms=True, indexOptions='docs_only')
         self.termsfilters = {}
+
     def termsfilter(self, filter, *others):
         "Return `TermsFilter`_ synced to given filter and optionally associated with other indexers."
         terms = self.sorter(self.field).terms(filter, *self.readers)
@@ -745,10 +856,12 @@ class ParallelIndexer(Indexer):
             termsfilter.refresh(other)
             other.termsfilters.add(termsfilter)
         return termsfilter
+
     def update(self, value, document=(), **terms):
         "Atomically update document based on unique field."
         terms[self.field] = value
         self.updateDocument(index.Term(self.field, value), self.document(document, **terms))
+
     def refresh(self, **caches):
         "Store refreshed searcher and synchronize :attr:`termsfilters`."
         sorter, segments = self.sorter(self.field), self.segments
