@@ -1,7 +1,8 @@
 from future_builtins import map, zip
 import unittest
 import os
-import tempfile, shutil
+import tempfile
+import shutil
 import itertools
 import warnings
 import datetime
@@ -16,11 +17,13 @@ from org.apache.pylucene.search import PythonFilter
 from lupyne import engine
 from . import fixture
 
+
 class typeAsPayload(engine.TokenFilter):
     "Custom implementation of lucene TypeAsPayloadTokenFilter."
     def setattrs(self):
         self.payload = self.type
 
+
 @contextlib.contextmanager
 def assertWarns(*categories):
     with warnings.catch_warnings(record=True) as messages:
@@ -28,19 +31,22 @@ def assertWarns(*categories):
     types = [message.category for message in messages]
     assert types == list(categories), types
 
+
 class Filter(PythonFilter):
     "Broken filter to test errors are raised."
     def getDocIdSet(self, *args):
         assert False
 
+
 class BaseTest(unittest.TestCase):
     def setUp(self):
         self.tempdir = tempfile.mkdtemp(dir=os.path.dirname(__file__))
     def tearDown(self):
         shutil.rmtree(self.tempdir)
 
+
 class TestCase(BaseTest):
-    
+
     def testInterface(self):
         "Indexer and document interfaces."
         self.assertRaises(TypeError, engine.IndexSearcher)
@@ -56,7 +62,7 @@ class TestCase(BaseTest):
             token.offset, token.positionIncrement = (0, 0), 0
         assert str(stemmer.parse('hellos', field=['body', 'title'])) == 'body:hello title:hello'
         assert str(stemmer.parse('hellos', field={'body': 1.0, 'title': 2.0})) == 'body:hello title:hello^2.0'
-        indexer = engine.Indexer(analyzer=stemmer, version=util.Version.LUCENE_43, writeLockTimeout=100L)
+        indexer = engine.Indexer(analyzer=stemmer, version=util.Version.LUCENE_46, writeLockTimeout=100L)
         assert indexer.config.writeLockTimeout == 100
         self.assertRaises(lucene.JavaError, engine.Indexer, indexer.directory)
         indexer.set('text')
@@ -174,7 +180,7 @@ class TestCase(BaseTest):
         self.assertRaises(AttributeError, getattr, reader, 'maxDoc')
         del indexer.indexSearcher
         self.assertRaises(AttributeError, getattr, indexer, 'search')
-    
+
     def testBasic(self):
         "Text fields and simple searches."
         self.assertRaises(lucene.JavaError, engine.Indexer, self.tempdir, 'r')
@@ -185,7 +191,7 @@ class TestCase(BaseTest):
             indexer.add(doc)
         indexer.commit()
         searcher = engine.IndexSearcher.load(self.tempdir)
-        engine.IndexSearcher.load(searcher.directory) # ensure directory isn't closed
+        engine.IndexSearcher.load(searcher.directory)  # ensure directory isn't closed
         assert len(indexer) == len(searcher) and store.RAMDirectory.instance_(searcher.directory)
         assert indexer.filters == indexer.spellcheckers == {}
         assert indexer.facets(engine.Query.alldocs(), 'amendment')
@@ -202,18 +208,6 @@ class TestCase(BaseTest):
         doc['date'] = engine.Analyzer(tokenizer).tokens(doc['date']), 2.0
         scores = list(searcher.match(doc, 'text:congress', 'text:law', 'amendment:27', 'date:19*'))
         assert 0.0 == scores[0] < scores[1] < scores[2] < scores[3] == 1.0
-        searcher = engine.MultiSearcher([indexer.indexReader, self.tempdir])
-        assert searcher.refCount == 1 and searcher.timestamp
-        assert searcher.count() == len(searcher) == 2 * len(indexer)
-        searcher.sorters['amendment'] = engine.SortField('amenmdment', int)
-        comparator = searcher.comparator('amendment')
-        assert set(map(type, comparator)) == set([int])
-        assert searcher is searcher.reopen()
-        facets = searcher.facets(engine.Query.alldocs(), 'amendment')['amendment']
-        assert facets.pop(None) == 16 and facets == dict.fromkeys(map(str, range(1, 28)), 2)
-        reader = searcher.indexReader
-        del searcher
-        assert not reader.refCount
         assert len(indexer) == len(indexer.search()) == 35
         assert sorted(indexer.names()) == ['amendment', 'article', 'date', 'text']
         articles = list(indexer.terms('article'))
@@ -250,9 +244,9 @@ class TestCase(BaseTest):
         sort = engine.SortField('amendment', type=int)
         hits = indexer.search('text:people', count=5, sort=sort)
         assert [hit.get('amendment') for hit in hits] == [None, None, '1', '2', '4']
-        assert [key for hit in hits for key in hit.keys]== [0, 0, 1, 2, 4]
+        assert [key for hit in hits for key in hit.keys] == [0, 0, 1, 2, 4]
         assert all(map(math.isnan, hits.scores))
-        hits = indexer.search('text:right', count=10**7, sort=sort, scores=True)
+        hits = indexer.search('text:right', count=10 ** 7, sort=sort, scores=True)
         assert not any(map(math.isnan, hits.scores)) and sorted(hits.scores, reverse=True) != hits.scores
         assert math.isnan(hits.maxscore)
         hits = indexer.search('text:right', count=2, sort=sort, maxscore=True)
@@ -364,7 +358,7 @@ class TestCase(BaseTest):
         assert not os.path.exists(os.path.join(self.tempdir, commit.segmentsFileName))
         assert engine.IndexWriter.check(self.tempdir).clean
         assert not engine.IndexWriter.check(self.tempdir, fix=True).numBadSegments
-    
+
     def testAdvanced(self):
         "Large data set with hierarchical fields."
         indexer = engine.Indexer(self.tempdir)
@@ -411,11 +405,10 @@ class TestCase(BaseTest):
         assert all(grouping.search(indexer.indexSearcher, engine.Query.alldocs()).facets.values())
         assert len(grouping) == len(list(grouping)) > 100
         assert set(grouping) > set(facets)
-        for count in (None, len(indexer)):
-            hits = indexer.search(query, count=count, timeout=0.01)
-            assert 0 <= len(hits) <= indexer.count(query) and hits.count in (None, len(hits)) and hits.maxscore in (None, 1.0)
-            hits = indexer.search(query, count=count, timeout=-1)
-            assert len(hits) == 0 and hits.count is hits.maxscore is None
+        hits = indexer.search(query, timeout=-1)
+        assert not hits and (hits.count is hits.maxscore is None)
+        hits = indexer.search(query, timeout=10)
+        assert len(hits) == hits.count == indexer.count(query) and hits.maxscore == 1.0
         self.assertRaises(lucene.JavaError, indexer.search, filter=Filter())
         directory = store.RAMDirectory()
         query = engine.Query.term('state', 'CA')
@@ -428,7 +421,7 @@ class TestCase(BaseTest):
         searcher = engine.IndexSearcher(path)
         assert len(searcher.segments) == 1 and 'CA' not in searcher.terms('state')
         directory.close()
-    
+
     def testSpatial(self):
         "Spatial tiles."
         indexer = engine.Indexer(self.tempdir, 'w')
@@ -461,28 +454,28 @@ class TestCase(BaseTest):
         hits = indexer.search(field.near(x, y, precision=10))
         cities = set(hit['city'] for hit in hits)
         assert city in cities and len(cities) > 10
-        query = field.within(x, y, 10**4)
+        query = field.within(x, y, 10 ** 4)
         assert len(query) < 3
         distances = indexer.distances(x, y, 'longitude', 'latitude')
         hits = indexer.search(query).sorted(distances.__getitem__)
         assert hits[0]['zipcode'] == zipcode and distances[hits[0].id] < 10
         cities = set(hit['city'] for hit in hits)
         assert city in cities and 100 > len(cities) > 50
-        hits = indexer.search(field.within(x, y, 10**5, limit=100))
+        hits = indexer.search(field.within(x, y, 10 ** 5, limit=100))
         cities = set(hit['city'] for hit in hits)
         assert city in cities and len(cities) > 100
-        ranges = 10**2, 10**5
+        ranges = 10 ** 2, 10 ** 5
         groups = hits.groupby(lambda id: bisect.bisect_left(ranges, distances[id]))
         counts = dict((hits.value, len(hits)) for hits in groups)
         assert 1 == counts[0] < counts[2] < counts[1]
-        assert len(field.within(x, y, 10**8)) == 1
+        assert len(field.within(x, y, 10 ** 8)) == 1
         self.assertRaises(NameError, list, field.radiate(y, x, 1, 0))
-        hits = hits.filter(lambda id: distances[id] < 10**4)
+        hits = hits.filter(lambda id: distances[id] < 10 ** 4)
         assert 0 < len(hits) < sum(counts.values())
         hits = hits.sorted(distances.__getitem__, reverse=True)
         ids = list(hits.ids)
-        assert 0 == distances[ids[-1]] < distances[ids[0]] < 10**4
-    
+        assert 0 == distances[ids[-1]] < distances[ids[0]] < 10 ** 4
+
     def testFields(self):
         "Custom fields."
         self.assertRaises(lucene.InvalidArgsError, engine.Field, '', stored='invalid')
@@ -495,10 +488,6 @@ class TestCase(BaseTest):
         field, = field.items(' ')
         attrs = 'indexed', 'tokenized', 'storeTermVectors', 'storeTermVectorPositions', 'storeTermVectorOffsets', 'omitNorms'
         assert all(getattr(field.fieldType(), attr)() for attr in attrs)
-        field, = engine.DocValuesField('', 'sorted').items('')
-        assert str(field.fieldType().docValueType()) == 'SORTED' and field.binaryValue().utf8ToString() == ''
-        field, = engine.DocValuesField('', 'numeric').items(0)
-        assert str(field.fieldType().docValueType()) == 'NUMERIC' and field.numericValue().longValue() == 0
         indexer = engine.Indexer(self.tempdir)
         indexer.set('amendment', engine.MapField, func='{0:02d}'.format, stored=True)
         indexer.set('size', engine.MapField, func='{0:04d}'.format, stored=True)
@@ -549,7 +538,7 @@ class TestCase(BaseTest):
         assert list(indexer.comparator('year'))[-1] == 0
         assert cache == len(search.FieldCache.DEFAULT.cacheEntries)
         self.assertRaises(AttributeError, indexer.comparator, 'size', type='score')
-    
+
     def testNumeric(self):
         "Numeric fields."
         nf, = engine.NumericField('temp').items(0.5)
@@ -575,7 +564,7 @@ class TestCase(BaseTest):
         assert [hit['amendment'] for hit in hits] == [18, 19]
         assert [datetime.datetime.utcfromtimestamp(float(hit['date'])).year for hit in hits] == [1919, 1920]
         assert indexer.count(field.within(seconds=100)) == indexer.count(field.within(weeks=1)) == 0
-        query = field.duration([2009], days=-100*365)
+        query = field.duration([2009], days=-100 * 365)
         assert indexer.count(query) == 12
         field = indexer.fields['size']
         assert len(list(indexer.terms('size'))) > len(indexer)
@@ -588,9 +577,9 @@ class TestCase(BaseTest):
         assert list(hits.ids) == ids[:len(hits)]
         query = field.range(None, 1000)
         assert indexer.count(query) == len(sizes) - len(ids)
-        self.assertRaises(OverflowError, list, field.items(-2**64))
-        assert str(field.range(-2**64, 0)) == 'size:[* TO 0}'
-        assert str(field.range(0, 2**64)) == 'size:[0 TO *}'
+        self.assertRaises(OverflowError, list, field.items(-2 ** 64))
+        assert str(field.range(-2 ** 64, 0)) == 'size:[* TO 0}'
+        assert str(field.range(0, 2 ** 64)) == 'size:[0 TO *}'
         assert str(field.range(0.5, None, upper=True)) == 'size:[0.5 TO *]'
         for step, count in zip(range(0, 20, field.numericPrecisionStep()), (26, 19, 3, 1)):
             sizes = list(indexer.numbers('size', step))
@@ -599,7 +588,7 @@ class TestCase(BaseTest):
             assert sum(numbers.values()) == len(indexer) and all(isinstance(number, float) for number in numbers)
         hit, = indexer.search(indexer.fields['amendment'].term(1))
         assert hit['amendment'] == 1
-    
+
     def testHighlighting(self):
         "Highlighting text fragments."
         indexer = engine.Indexer()
@@ -618,7 +607,7 @@ class TestCase(BaseTest):
         query = '"persons, houses, papers"'
         highlighter = indexer.highlighter(query, '', terms=True, fields=True, formatter=highlight.SimpleHTMLFormatter('*', '*'))
         fragments = highlighter.fragments(text, count=3)
-        assert len(fragments) == 2 and fragments[0].count('*') == 2*3 and '*persons*' in fragments[1]
+        assert len(fragments) == 2 and fragments[0].count('*') == 2 * 3 and '*persons*' in fragments[1]
         highlighter = indexer.highlighter(query, '', terms=True)
         highlighter.textFragmenter = highlight.SimpleFragmenter(200)
         fragment, = highlighter.fragments(text, count=3)
@@ -630,10 +619,10 @@ class TestCase(BaseTest):
         highlighter = indexer.highlighter(query, 'text', fragListBuilder=vectorhighlight.SingleFragListBuilder())
         text, = highlighter.fragments(id)
         assert fragment in text and len(text) > len(fragment)
-    
+
     def testNearRealTime(self):
         "Near real-time index updates."
-        indexer = engine.Indexer(version=util.Version.LUCENE_43, nrt=True)
+        indexer = engine.Indexer(version=util.Version.LUCENE_46, nrt=True)
         indexer.add()
         assert indexer.count() == 0 and not indexer.current
         indexer.refresh(filters=True)
@@ -643,26 +632,32 @@ class TestCase(BaseTest):
         indexer.add()
         indexer.commit()
         assert indexer.count() == engine.IndexSearcher(indexer.directory).count() == 2
-    
+
     def testFilters(self):
         "Custom filters."
         indexer = engine.Indexer()
-        indexer.set('name', stored=True, indexed=True)
-        for name in ('alpha', 'bravo', 'charlie'):
+        indexer.set('name', stored=True, tokenized=False)
+        for name in ('alpha', 'bravo'):
             indexer.add(name=name)
         indexer.commit()
-        filter = engine.TermsFilter('name')
-        assert len(filter.readers) == 0
+        filter = indexer.termsfilter('name')
+        assert filter.readers
+        indexer.add(name='charlie')
+        indexer.commit()
         filter.add('alpha', 'bravo')
         filter.discard('bravo', 'charlie')
         assert filter.values == set(['alpha'])
+        assert [hit['name'] for hit in indexer.search(filter=filter)] == ['alpha']
+        indexer.update('name', document={'name': 'alpha'})
+        indexer.commit()
+        assert [hit['name'] for hit in indexer.search(filter=filter)] == ['alpha']
         parallel = engine.ParallelIndexer('name')
         parallel.set('priority', tokenized=False)
         for name in ('alpha', 'bravo', 'delta'):
             parallel.update(name, priority='high')
         parallel.commit()
         filter = parallel.termsfilter(engine.Query.term('priority', 'high').filter(), indexer)
-        assert [hit['name'] for hit in indexer.search(filter=filter)] == ['alpha', 'bravo']
+        assert [hit['name'] for hit in indexer.search(filter=filter)] == ['bravo', 'alpha']
         indexer.add(name='delta')
         indexer.delete('name', 'alpha')
         indexer.commit()
@@ -675,6 +670,51 @@ class TestCase(BaseTest):
         parallel.commit()
         filter.refresh(indexer)
         assert filter.readers == set(indexer.readers)
+        parallel.update('charlie', priority='low')
+        parallel.commit()
+        assert [hit['name'] for hit in indexer.search(filter=filter)] == ['delta']
+
+    def testMulti(self):
+        "MultiSearchers."
+        indexers = engine.Indexer(self.tempdir), engine.Indexer()
+        searcher = engine.MultiSearcher([indexers[0].indexReader, indexers[1].directory])
+        self.assertRaises(TypeError, getattr, searcher, 'timestamp')
+        assert engine.MultiSearcher(indexers[:1]).timestamp
+        assert [reader.refCount for reader in searcher.indexReaders] == [2, 1]
+        assert searcher.reopen() is searcher
+        indexers[0].add()
+        indexers[0].commit()
+        assert [reader.refCount for reader in searcher.indexReaders] == [1, 1]
+        searcher, previous = searcher.reopen(), searcher
+        assert searcher.version > previous.version
+        assert [reader.refCount for reader in searcher.indexReaders] == [1, 2]
+        del previous
+        assert [reader.refCount for reader in searcher.indexReaders] == [1, 1]
+
+    def testDocValues(self):
+        "DocValues and updates."
+        indexer = engine.Indexer()
+        indexer.set('id')
+        indexer.set('votes', engine.DocValuesField, type='numeric')
+        indexer.set('tag', engine.DocValuesField, type='binary')
+        indexer.add(id='0', votes=1, tag='low')
+        indexer.commit()
+        segments = indexer.segments
+        indexer.update('id', id='0', votes=2, tag='medium')
+        indexer.commit()
+        assert indexer.segments != segments
+        segments = indexer.segments
+        assert list(indexer.comparator('votes', type=int)) == [2]
+        assert list(indexer.comparator('tag', type='bytes')) == ['medium']
+        indexer.update('id', '0', votes=3)
+        indexer.commit()
+        assert indexer.segments == segments
+        assert list(indexer.comparator('votes', type=int)) == [3]
+        if lucene.VERSION >= '4.8':
+            indexer.update('id', '0', tag='high')
+            indexer.commit()
+            assert indexer.segments == segments
+            assert list(indexer.comparator('tag', type='bytes')) == ['high']
 
 if __name__ == '__main__':
     lucene.initVM()
