@@ -1,14 +1,18 @@
 from future_builtins import map
 import unittest
-import os, sys
-import tempfile, shutil
-import signal, subprocess
+import os
+import sys
+import tempfile
+import shutil
+import signal
+import subprocess
 import operator
 import httplib
 import math
 import json
-import time, calendar
-import socket, errno
+import time
+import calendar
+import errno
 import contextlib
 from email.utils import parsedate
 import lucene
@@ -16,6 +20,7 @@ import cherrypy
 from lupyne import client, engine, server
 from . import fixture, local
 
+
 @contextlib.contextmanager
 def assertRaises(exception, code):
     "Assert an exception is raised with specific code."
@@ -26,6 +31,7 @@ def assertRaises(exception, code):
     else:
         raise AssertionError(exception.__name__ + ' not raised')
 
+
 class Resource(client.Resource):
     "Modify status and inject headers to test warning framework."
     def getresponse(self):
@@ -36,12 +42,16 @@ class Resource(client.Resource):
             response.msg.addheader('location', 'http://{0}:{1}'.format(self.host, self.port))
         return response
 
+
 class BaseTest(local.BaseTest):
+    module = 'lupyne.server'
     ports = 8080, 8081
     config = {'server.socket_timeout': 2, 'server.shutdown_timeout': 1}
+
     def run(self, result):
         self.config['log.screen'] = result.showAll
         local.BaseTest.run(self, result)
+
     def setUp(self):
         local.BaseTest.setUp(self)
         self.servers = {}
@@ -49,12 +59,13 @@ class BaseTest(local.BaseTest):
         for port in list(self.servers):
             self.stop(port)
         local.BaseTest.tearDown(self)
+
     def start(self, port, *args, **config):
         "Start server in separate process on given port."
         config.update(self.config)
         config['server.socket_port'] = port
         cherrypy.process.servers.wait_for_free_port('localhost', port)
-        server = self.servers[port] = subprocess.Popen((sys.executable, '-m', 'lupyne.server', '-c', json.dumps(config)) + args)
+        server = self.servers[port] = subprocess.Popen((sys.executable, '-m', self.module, '-c', json.dumps(config)) + args)
         cherrypy.process.servers.wait_for_occupied_port('localhost', port)
         assert not server.poll()
     def stop(self, port):
@@ -63,13 +74,14 @@ class BaseTest(local.BaseTest):
         server.terminate()
         assert server.wait() == 0
 
+
 class TestCase(BaseTest):
-    
+
     def testInterface(self):
         "Remote reading and writing."
         config = {'tools.json_out.indent': 2, 'tools.validate.last_modified': True, 'tools.validate.expires': 0, 'tools.validate.max_age': 0}
         self.start(self.ports[0], self.tempdir, '--autoreload=1', **config),
-        self.start(self.ports[1], self.tempdir, self.tempdir, '--autoupdate=2.0'), # concurrent searchers
+        self.start(self.ports[1], self.tempdir, self.tempdir, '--autoupdate=2.0'),  # concurrent searchers
         resource = client.Resource('localhost', self.ports[0])
         assert resource.get('/favicon.ico')
         response = resource.call('GET', '/')
@@ -200,7 +212,7 @@ class TestCase(BaseTest):
         assert 0 < result['maxscore'] < 1
         doc, = result['docs']
         assert sorted(doc) == ['__id__', '__score__', 'name']
-        assert doc['__id__'] == 0 and doc['__score__'] > 0 and doc['name'] == 'sample' 
+        assert doc['__id__'] == 0 and doc['__score__'] > 0 and doc['name'] == 'sample'
         hit, = resource.get('/search', q='hello world', **{'q.field': ['text', 'body']})['docs']
         assert hit['__id__'] == doc['__id__'] and hit['__score__'] < doc['__score__']
         hit, = resource.get('/search', q='hello world', **{'q.field': 'text', 'q.op': 'and'})['docs']
@@ -272,7 +284,7 @@ class TestCase(BaseTest):
         with assertRaises(httplib.HTTPException, httplib.NOT_FOUND):
             assert resource.put('/update/backup') == names
         resource = client.Resource('localhost', self.ports[-1] + 1)
-        with assertRaises(socket.error, errno.ECONNREFUSED):
+        with assertRaises(IOError, errno.ECONNREFUSED):
             resource.get('/')
         port = self.ports[0]
         self.stop(port)
@@ -286,7 +298,7 @@ class TestCase(BaseTest):
         assert cherrypy.tree.mount(None)
         server.init(vmargs=None)
         self.assertRaises(AttributeError, server.start, config=True)
-    
+
     def testBasic(self):
         "Remote text indexing and searching."
         self.start(self.ports[0], self.tempdir)
@@ -424,7 +436,7 @@ class TestCase(BaseTest):
         assert set(map(operator.itemgetter('count'), result['groups'])) == set([1])
         assert all(int(doc.get('amendment', 0)) == group['value'] for group in result['groups'] for doc in group['docs'])
         assert result['groups'][0]['value'] == 2 and result['groups'][-1]['value'] == 0
-    
+
     def testAdvanced(self):
         "Nested and numeric fields."
         writer = engine.IndexWriter(self.tempdir)
@@ -440,8 +452,8 @@ class TestCase(BaseTest):
         assert resource.post('/update') == resource.get('/').popitem()[1] == len(writer)
         terms = resource.get('/terms/zipcode:int')
         assert len(terms) == len(writer) and terms[0] == 90001
-        terms = resource.get('/terms/zipcode:int?step=4')
-        assert len(terms) < len(writer) and terms[0] == 90000
+        terms = resource.get('/terms/zipcode:int?step=16')
+        assert terms == [65536]
         result = resource.get('/search', count=0, facets='county')
         facets = result['facets']['county']
         assert result['count'] == sum(facets.values()) and 'Los Angeles' in facets
@@ -472,7 +484,7 @@ class TestCase(BaseTest):
         assert resource.delete('/queries/default/alpha') == 'name:alpha'
         assert resource.delete('/queries/default/alpha') is None
         assert resource.get('/queries/default') == {'bravo': 0.0}
-    
+
     def testRealTime(self):
         "Real Time updating and searching."
         for args in [('-r',), ('--real-time', 'index0', 'index1'), ('-r', '--real-time', 'index')]:
@@ -499,6 +511,23 @@ class TestCase(BaseTest):
         response = resource.call('GET', '/docs')
         assert response and version != response.getheader('etag')
 
+    def example(self):
+        "Custom server example (only run explicitly)."
+        self.module = 'examples.server'
+        port = self.ports[0]
+        self.start(port)
+        resource = client.Resource('localhost', port)
+        result = resource.get('/search', q='date:17*', group='year')
+        assert dict(map(operator.itemgetter('value', 'count'), result['groups'])) == {1795: 1, 1791: 10}
+        result = resource.get('/search', q='date:17*', group='year', sort='-year')
+        assert list(map(operator.itemgetter('value'), result['groups'])) == [1795, 1791]
+        result = resource.get('/search', count=0, facets='year')
+        facets = result['facets']['year']
+        assert not result['docs'] and facets['1791'] == 10 and sum(facets.values()) == result['count']
+        result = resource.get('/search', q='text:right', facets='year')
+        facets = result['facets']['year']
+        assert len(result['docs']) == result['count'] == sum(facets.values())
+
 if __name__ == '__main__':
     lucene.initVM()
     unittest.main()
